# Take the character & bitmap data from SmallVector.txt, generate a bitmap image
# for each character, trace that image and generate FontForge spline data.
#
# Jason Hood, 8 to 17 October, 2014.
# 25 August, 2015: generate "uniXXXX" names automatically.
# 29 August to 1 September, 2015: work with the combined UCS fonts.
#
# According to FontForge, TrueType fonts prefer a power of 2 size, so we'll use
# 2048 - that makes each pixel 2048/10 = 204 em, leaving a margin of 4 top and
# bottom.  However, use 222 em for the width, in order to preserve the width
# relative to the height.
#
# sml2sfd < SmallVector.txt > SmallVector.sfd

($WIDTH, $ASCENT) = (1332, 1632);

get_tracer();

$now = time;

# Write the SFD header and info.
  print <<END;
SplineFontDB: 3.0
FontName: SmallVector
FullName: SmallVector
FamilyName: SmallVector
Weight: Regular
Copyright: Public domain font.  Share and enjoy.
Version: 1.0
ItalicAngle: 0
UnderlinePosition: 0
UnderlineWidth: 0
Ascent: 1636
Descent: 412
LayerCount: 2
Layer: 0 1 "Back"  1
Layer: 1 1 "Fore"  0
XUID: [1021 295 18731 817]
FSType: 8
OS2Version: 0
OS2_WeightWidthSlopeOnly: 0
OS2_UseTypoMetrics: 1
CreationTime: $now
ModificationTime: $now
PfmFamily: 49
TTFWeight: 400
TTFWidth: 5
LineGap: 0
VLineGap: 0
Panose: 2 0 5 9 0 0 0 0 0 0
OS2TypoAscent: 0
OS2TypoAOffset: 1
OS2TypoDescent: 0
OS2TypoDOffset: 1
OS2TypoLinegap: 0
OS2WinAscent: 0
OS2WinAOffset: 1
OS2WinDescent: 0
OS2WinDOffset: 1
HheadAscent: 0
HheadAOffset: 1
HheadDescent: 0
HheadDOffset: 1
OS2FamilyClass: 2057
OS2Vendor: 'PfEd'
DEI: 91125
LangName: 1033 "" "" "" "" "" "" "" "" "" "" "" "http://terminalvector.adoxa.vze.com/"
GaspTable: 2 10 1 65535 10 1
Encoding: UnicodeBmp
Compacted: 1
UnicodeInterp: none
NameList: Adobe Glyph List
DisplaySize: -20
AntiAlias: 0
FitToEm: 1
WinInfo: 0 32 16
BeginChars: 65536

END

$update = time;

while (<STDIN>) {
  next if /^\s*(;|$)/;

  # Write the initial character info.
  my ($char, $name) = split(/\s/);
  $name = 'uni' . substr($char, 2) unless $name =~ /^\w/;
  $code = hex(substr($char, 2));
  print <<END;
StartChar: $name
Encoding: $code $code
Width: $WIDTH
Flags: W
LayerCount: 2
Fore
END

  # Turn the text bitmap into a PBM image, scaled up 3x (the minimum potrace
  # needs to keep the right angles).
  # Map each bitmap, to find duplicates and create references.
  open(PBM, ">glyph.pbm") || die "can't create glyph.pbm";
  binmode PBM;
  $bits = $bm10 = $bm9 = $bm8 = $bm7 = $bm6 = '';
  for ($i = 0;; ++$i) {
    $_ = <STDIN>;
    chomp;
    last if /^$/;
    tr/.#/01/;
    $_ = sprintf("%-42s", $_);
    ($line10, $line9, $line8, $line7, $line6) = unpack 'A6 x4 A6 x4 A5 x4 A5 x4 A4';
    if ($line6 =~ /^[01]/) {
      $line = $line6;
      $bm6 .= pack 'B4', $line;
      ($width, $height, $descent) = (4,  6, 1);
    }
    if ($line7 =~ /^[01]/) {
      $line = $line7;
      $bm7 .= pack 'B5', $line;
      ($width, $height, $descent) = (5,  7, 1);
    }
    if ($line8 =~ /^[01]/) {
      $line = $line8;
      $bm8 .= pack 'B5', $line;
      ($width, $height, $descent) = (5,  8, 1);
    }
    if ($line9 =~ /^[01]/) {
      $line = $line9;
      $bm9 .= pack 'B6', $line;
      ($width, $height, $descent) = (6,  9, 2);
    }
    if ($line10 =~ /^[01]/) {
      $line = $line10;
      $bm10 .= pack 'B6', $line;
      ($width, $height, $descent) = (6, 10, 2);
    }
    $bits .= pack "B$width", $line;
    $line =~ s/(.)/\1\1\1/g;
    if ($i == 0) {
      print PBM 'P4 ', $width * 3, ' ', $height * 3, "\n";
    }
    print PBM pack('B' . $width * 3, $line) x 3;
  }
  close PBM;

  push @code,  $code;
  push @bdf6,  $bm6;
  push @bdf7,  $bm7;
  push @bdf8,  $bm8;
  push @bdf9,  $bm9;
  push @bdf10, $bm10;

  if ($chars{$bits}) {
    print "Refer: $chars{$bits} N 1 0 0 1\n";
  }
  else {
    $chars{$bits} = $glyphs . ' ' . $code;
    print "SplineSet\n";
    # The tracer generates anticlockwise points for filled objects, whereas
    # FontForge wants clockwise - reverse each path (not that it really seems
    # to matter).
    @points = ();
    my $start = 0;
    if ($autotrace) {
      open(OUTLINE, "$tracer -background FFFFFF -corner-surround 1 -output-format epd glyph.pbm |");
      while (<OUTLINE>) {
	next if /^%|rg$/;
	($x, $y, $c) = split;
	if (($c eq 'm' and $#points != $start) or /f/) {
	  @points[$start+1..$#points-1] = reverse @points[$start+1..$#points-1];
	  last if /f/;
	  $start = @points;
	}
	push @points, int($x / 3 * $WIDTH / $width) . ' ' . int(($y / 3 - $descent) * $ASCENT / ($height - $descent)) . " $c 1\n";
      }
    }
    else {
      open(OUTLINE, "$tracer --alphamax 0 --cleartext --unit 0.33 --output - glyph.pbm |");
      while (<OUTLINE>) {
	if (/ i$/) {
	  $start = @points;
	  ($ox, $oy, $x, $y) = split;
	  push @points, int($ox * $WIDTH / $width) . ' ' . int(($oy - $descent) * $ASCENT / ($height - $descent)) . " m 1\n";
	  $ox += $x; $oy += $y;
	  push @points, int($ox * $WIDTH / $width) . ' ' . int(($oy - $descent) * $ASCENT / ($height - $descent)) . " l 1\n";
	}
	elsif (/ v$/) {
	  ($x, $y) = split;
	  $ox += $x; $oy += $y;
	  push @points, int($ox * $WIDTH / $width) . ' ' . int(($oy - $descent) * $ASCENT / ($height - $descent)) . " l 1\n";
	}
	elsif (/^v$/) {
	  @points[$start+1..$#points-1] = reverse @points[$start+1..$#points-1];
	}
      }
    }
    close OUTLINE;

    # FontForge complains about intersecting lines - move points by one em to
    # compensate (but again, it doesn't seem to matter).
    %point = ();
    $start = 0;
    for ($i = 1; $i < @points; ++$i) {
      # The last point of the path is supposed to be the same as the first
      # (just in case it's moved).
      if ($i == $#points or $points[$i+1] =~ /m 1$/) {
	$points[$i] = $points[$start] =~ tr/m/l/r;
	$start = $i + 1;
	next;
      }
      my $p = substr($points[$i], 0, -5);
      if ($point{$p}) {
	($x, $y, $c) = split(" ", $points[$i]);
	my ($x2, $y2) = split(" ", $points[$i + 1]);
	   if ($x > $x2) { --$x; }
	elsif ($x < $x2) { ++$x; }
	elsif ($y > $y2) { --$y; }
	elsif ($y < $y2) { ++$y; }
	$points[$i] = $x . ' ' . $y . " $c 1\n";
      }
      else {
	$point{$p} = 1;
      }
    }
    print join "", @points;
    print "EndSplineSet\n";
  }
  print "EndChar\n\n";
  ++$glyphs;

  if (time != $update) {
    print STDERR "\rGenerating glyphs... $glyphs";
    $update = time;
  }
}
unlink 'glyph.pbm';
print STDERR "\r", ' ' x 26, "\r";

print "EndChars\n\n";

write_bitmap(\@bdf6,  4,  6, 1);
write_bitmap(\@bdf7,  5,  7, 1);
write_bitmap(\@bdf8,  5,  8, 1);
write_bitmap(\@bdf9,  6,  9, 2);
write_bitmap(\@bdf10, 6, 10, 2);

print "EndSplineFont\n";


sub write_bitmap {
  my ($bref, $width, $height, $descent) = @_;
  print "BitmapFont: $height 65536 ", $height - $descent, " $descent 1\n";
  for (my $i = 0; $i < $glyphs; ++$i) {
    write_bdf($i, @$bref[$i], $width, $descent) if @$bref[$i];
  }
  print "EndBitmapFont\n\n";
}


sub write_bdf {
  my $left, $right, $bot, $top;
  my $i, $horz, $num;
  my ($glyph, $bits, $width, $descent) = @_;
  my $last = length($bits) - 1;
  # Ensure the last line can be expanded to a long for the base85 conversion.
  my @lines = unpack('C*', $bits . "\0\0\0");

  print "BDFChar: $glyph $code[$glyph] $width ";
  $bot = $last;
  while ($lines[$bot] == 0 and $bot >= 0) { --$bot; }
  if ($bot == -1) {
    print "0 0 0 0\nz\n";
    return;
  }
  $top = 0;
  while ($lines[$top] == 0) { ++$top; }
  $horz = 0;
  for ($i = $top; $i <= $bot; ++$i) { $horz |= $lines[$i]; }
  $left = 0;
  $i = 0x80;
  while (($horz & $i) == 0) { ++$left; $i >>= 1; }
  $right = 7;
  $i = 1;
  while (($horz & $i) == 0) { --$right; $i <<= 1; }
  # 5x7 & 5x8 move up a line for some reason, so shift down to compensate.
  --$last if $last == 6 or $last == 7;
  print "$left $right ", $last - $bot - $descent, ' ', $last - $top - $descent, "\n";
  for ($i = $top; $i <= $bot; $i += 4) {
    $num  = $lines[$i + 0] << $left << 24;
    $num |= $lines[$i + 1] << $left << 16;
    $num |= $lines[$i + 2] << $left << 8;
    $num |= $lines[$i + 3] << $left << 0;
    print base85($num);
  }
  print "\n";
}

sub base85 {
  my $num = $_[0];
  return "z" if $num == 0;

  my $str = chr(($num % 85) + ord('!'));
  $num = int($num / 85);
  $str = chr(($num % 85) + ord('!')) . $str;
  $num = int($num / 85);
  $str = chr(($num % 85) + ord('!')) . $str;
  $num = int($num / 85);
  $str = chr(($num % 85) + ord('!')) . $str;
  $num = int($num / 85);
  $str = chr(($num % 85) + ord('!')) . $str;

  return $str;
}


sub get_tracer {
  $autotrace = 1;
  $tracer = $ENV{'AUTOTRACE'} || 'autotrace';
  open(my $fh, "$tracer -version 2>nul |");
  if (not <$fh>) {
    $autotrace = 0;
    $tracer = $ENV{'POTRACE'} || 'potrace';
    open($fh, "$tracer -version 2>nul |");
    die "AutoTrace or potrace required" unless <$fh>;
  }
  close $fh;
}
